#include <OneWire.h>
#include <DallasTemperature.h>
#include <EEPROM.h>
#include <PID_v1.h>
#include <PID_AutoTune_v0.h>
#include <MemoryFree.h>
#include"Default.h"
#include <avr/pgmspace.h>

#define DEBUG 1                       // incl. seriellem Monitor zum debuggen
#define SONIC 1                       // incl. Code für Ultraschall-Sensor

/* **********************************************************************************************


     ######    ###   ######           #####
     #     #    #    #     #         #     #   ####   #    #   #####  #####    ####    #
     #     #    #    #     #         #        #    #  ##   #     #    #    #  #    #   #
     ######     #    #     #  #####  #        #    #  # #  #     #    #    #  #    #   #
     #          #    #     #         #        #    #  #  # #     #    #####   #    #   #
     #          #    #     #         #     #  #    #  #   ##     #    #   #   #    #   #
     #         ###   ######           #####    ####   #    #     #    #    #   ####    ######

                                 #    #    ##    #    #   ####
                                 ##   #   #  #   ##   #  #    #
                                 # #  #  #    #  # #  #  #    #
                                 #  # #  ######  #  # #  #    #
                                 #   ##  #    #  #   ##  #    #
                                 #    #  #    #  #    #   ####

   *
   * Author    : H.Herrmannsdörfer
   * Created   : 11/2014
   * Decription: Auto-Calculate PID parameters for the BrewControl
   * Changes   : 24.11.2014 V1 see https://learn.adafruit.com/sous-vide-powered-by-arduino-the-sous-viduino/the-whole-enchilada
   *
   ****************************************************************************/

// global variables
const int WINDOWSIZE = 5000; 
byte            relaisstate = 0;           // akt. Status der Relais-Ausgaenge (bitkodiert , RELAIS1 = 0, RELAIS2=1, ...)
float           akttempsC[4] = { -50.0, -50.0, -50.0, -50.0 };
unsigned long   postingInterval = WINDOWSIZE;  // delay between updates, in milliseconds

/*-----( Declare objects )-----*/
OneWire oneWire(SENSORPIN);                             // Setup a oneWire instance to communicate with any OneWire devices
DallasTemperature sensors(&oneWire);                    // Pass our oneWire reference to Dallas Temperature. 

// Local settings
SettingsData  settings  = {
    1,
    { 192, 168, 100, 150 },
    { 192, 168, 100, 1   },
    1,
    0,
    { 
      { 0x28, 0x04, 0xD5, 0x4D, 0x04, 0x00, 0x00, 0xAD }, 
      { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
      { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
      { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
    },
    {   0,   0,   0,   0 },
    850.0,
    0.5,
    0.1
  };

double pidSet, pidInput, pidOutput;
volatile long onTime = 0;            // PID Ergebnis
PID myPID = PID(&pidInput, &pidOutput, &pidSet,0.0,0.0,0.0, DIRECT);
// 10 second Time Proportional Output window
unsigned long windowStartTime;

// ************************************************
// Auto Tune Variables and constants
// ************************************************
byte         ATuneModeRemember=2;
double       aTuneStep=101;
double       aTuneNoise=.5;
unsigned int aTuneLookBack=20;
boolean      tuning = false;
 
PID_ATune aTune(&pidInput, &pidOutput);

const int logInterval = 10000; // log every 10 seconds
long lastLogTime = 0;

// ================================================
// SETUP
// ================================================
void setup(void) {
  uint8_t epromtest;

  Serial.begin(9600);
  Serial.println("PIDControl ... setup");

  // have the settings ever been written ? read the first byte of the settings to check 
  // that. otherwise take the default values
  epromtest = EEPROM.read(0);
  if (epromtest != 0 && epromtest != 255) {
    EEPROM_readAnything(0, settings);
    Serial.println("Read settings from EEPROM...");
  } else {
    resetValues();
  }
  resetValues();
  
  double aTuneStep=250, aTuneNoise=0.3, aTuneStartValue=1000;
  
  // setup known sensors
  if (settings.Sensors[0][0] != 0) sensors.setResolution(settings.Sensors[0], 12); 
  
  // Relais-Pins einstellen
  pinMode(RELAYPIN_1, OUTPUT);
  pinMode(RELAYPIN_2, OUTPUT);
  pinMode(RELAYPIN_3, OUTPUT);
  digitalWrite(RELAYPIN_1, RELAY_OFF);
  digitalWrite(RELAYPIN_2, RELAY_OFF);
  digitalWrite(RELAYPIN_3, RELAY_OFF);

  // PID initialisieren
  myPID.SetTunings(settings.Kp, settings.Ki, settings.Kd);
  myPID.SetSampleTime(1000);
  myPID.SetOutputLimits(0, WINDOWSIZE);
  myPID.SetMode(AUTOMATIC);             // PID anschalten
  
  pidSet = 66.0;                        // irgendeine mittlerere Maischtemperatur ... 
  
  aTune.SetControlType(1);
  
  StartAutoTune();
  lastLogTime = millis() + logInterval;
}



void loop(void) {
  /* ===========================================================
  Bekannte Temperaturen auslesen
  =========================================================== */
  getTemperatures(); 
  pidInput = akttempsC[0];

  if (tuning) { // run the auto-tuner
    if (aTune.Runtime()) { // returns 'true' when done
      FinishAutoTune();
    }
    onTime = pidOutput;
  } else {
    onTime = 0;
  } 

  DriveOutput();

  // periodically log to serial port in csv format
  if (millis() - lastLogTime > logInterval) {
    lastLogTime = millis();
    Serial.print("Auto On : ");
    Serial.print(tuning);
    Serial.print("  Set : ");
    Serial.print(pidSet);
    Serial.print("  Input : ");
    Serial.print(pidInput);
    Serial.print(" => ");
    Serial.println(pidOutput);
  }
  delay(100);
}








/* ===========================================================
   getTemperatures
   
   Temperaturen von allen bekannten Sensoren abfragen.
   =========================================================== */
void getTemperatures() {
  sensors.requestTemperatures();
  if (settings.Sensors[0][0] != 0) akttempsC[0] = sensors.getTempC(settings.Sensors[0]);
  if (settings.Sensors[1][0] != 0) akttempsC[1] = sensors.getTempC(settings.Sensors[1]); 
  if (settings.Sensors[2][0] != 0) akttempsC[2] = sensors.getTempC(settings.Sensors[2]); 
  if (settings.Sensors[3][0] != 0) akttempsC[3] = sensors.getTempC(settings.Sensors[3]); 
}
 

// ************************************************
// Start the Auto-Tuning cycle
// ************************************************
 
void StartAutoTune() {
  // set up the auto-tune parameters
  aTune.SetNoiseBand(aTuneNoise);
  aTune.SetOutputStep(aTuneStep);
  aTune.SetLookbackSec((int)aTuneLookBack);
  tuning = true;
}
 
// ************************************************
// Return to normal control
// ************************************************
void FinishAutoTune() {
  char string_temp[3][8] = {"0.0","0.0","0.0"};

  tuning = false;
 
  // Extract the auto-tune calculated parameters
  dtostrf(aTune.GetKp(), 4, 1, string_temp[0]);    
  dtostrf(aTune.GetKi(), 4, 1, string_temp[1]);    
  dtostrf(aTune.GetKd(), 4, 1, string_temp[2]);    

  settings.Kp = aTune.GetKp();
  settings.Ki = aTune.GetKi();
  settings.Kd = aTune.GetKd();
 
  Serial.print("P : ");
  Serial.print(string_temp[0]);
  Serial.print(" / I : ");
  Serial.print(string_temp[1]);
  Serial.print(" / D : ");
  Serial.print(string_temp[2]);
}
 
// ===========================================================
// resetValues
// 
// Alle Einstellungswerte in den Grundzustand zurücksetzen. Wird
// auch beim Setup aufgerufen, wenn nichts im EEPROM gespeichert 
// war.
// =========================================================== 
void resetValues() {
  settings.anl_no        = 1;
  settings.heatercount   = 1;
  settings.max_water     = 0;
  
  // Sensor addressen
  settings.Sensors[0][0] = 0x28;  settings.Sensors[0][1] = 0x04;  settings.Sensors[0][2] = 0xD5;  settings.Sensors[0][3] = 0x4D;
  settings.Sensors[0][4] = 0x04;  settings.Sensors[0][5] = 0x00;  settings.Sensors[0][6] = 0x00;  settings.Sensors[0][7] = 0xAD;

  settings.Sensors[1][0] = 0;  settings.Sensors[2][0] = 0;  settings.Sensors[3][0] = 0;
  
  // Sensor-Abweichungen
  settings.deviation[0]  = 0;  settings.deviation[1]  = 0;  settings.deviation[2]  = 0;  settings.deviation[3]  = 0;
  
  settings.src_ip[0] = 192;  settings.src_ip[1] = 168;  settings.src_ip[2] = 100;  settings.src_ip[3] = 150;
  settings.dst_ip[0] = 192;  settings.dst_ip[1] = 168;  settings.dst_ip[2] = 100;  settings.dst_ip[3] = 1;
  
  // PID-parameter für einen einfachen Einkocher 
  settings.Kp = 850;
  settings.Ki = 0.5;
  settings.Kd = 0.1;
}



// ************************************************
// Called by loop to drive the output
// ************************************************
void DriveOutput() {  
  long now = millis();
  
  // Set the output
  // "on time" is proportional to the PID output
  if ((now - windowStartTime) > WINDOWSIZE) { //time to shift the Relay Window
     windowStartTime += WINDOWSIZE;
  }
  if((onTime > 500) && (onTime > (now - windowStartTime))) {
     digitalWrite(RELAYPIN_1,RELAY_ON);
     if (settings.heatercount > 1) digitalWrite(RELAYPIN_2,RELAY_ON);
     if (settings.heatercount > 2) digitalWrite(RELAYPIN_3,RELAY_ON);
  } else {
     digitalWrite(RELAYPIN_1,RELAY_OFF);
     if (settings.heatercount > 1) digitalWrite(RELAYPIN_2,RELAY_OFF);
     if (settings.heatercount > 2) digitalWrite(RELAYPIN_3,RELAY_OFF);
  }
}

